package de.fraunhofer.iem.dataprovider.repository.service

import de.fraunhofer.iem.dataprovider.gitlab.service.OpenCodeGitlabApi
import de.fraunhofer.iem.dataprovider.kpi.dto.RawValueKpiCreateDto
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.dto.RepositoryConsentDto
import de.fraunhofer.iem.dataprovider.repository.dto.RepositoryCreateDto
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryEntity
import de.fraunhofer.iem.dataprovider.repository.repository.RepositoryRepository
import de.fraunhofer.iem.dataprovider.tool.dto.CreateToolDto
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import de.fraunhofer.iem.dataprovider.toolRun.dto.ToolRunDto
import de.fraunhofer.iem.dataprovider.toolRun.entity.LanguageEntity
import de.fraunhofer.iem.dataprovider.toolRun.entity.ToolRunEntity
import de.fraunhofer.iem.dataprovider.toolRun.repository.ToolRunRepository
import org.springframework.stereotype.Service
import org.springframework.transaction.annotation.Propagation
import org.springframework.transaction.annotation.Transactional

@Service
class RepositoryService(
    private val repositoryRepository: RepositoryRepository,
    private val openCodeGitlabApi: OpenCodeGitlabApi,
    private val toolService: ToolService,
    private val toolRunRepository: ToolRunRepository
) {
    private val logger = getLogger(javaClass)

    @Transactional(readOnly = false, propagation = Propagation.REQUIRES_NEW)
    fun createToolRun(
        repoDto: RepositoryCreateDto,
        kpiToolList: List<Pair<CreateToolDto, List<RawValueKpiCreateDto>>>,
        languageMap: Map<String, Float>
    ) {
        val repo = getOrCreate(repoDto)
        val toolRun = ToolRunEntity(repository = repo)

        val languages =
            languageMap.map {
                LanguageEntity(
                    toolRunEntity = toolRun,
                    name = it.key,
                    percentage = it.value
                )
            }
        toolRun.languageEntities.addAll(languages)

        kpiToolList.forEach { toolKpi ->
            toolService.findToolByName(toolKpi.first)?.let { tool ->
                toolRun.toolEntities.add(tool)
            }

            val kpis = toolKpi.second.map { rawKpi ->
                rawKpi.toDbObject(toolRun)
            }
            toolRun.kpiEntities.addAll(kpis)
        }

        repo.toolRuns.add(toolRun)
        repositoryRepository.save(repo)
    }

    /**
     * Either creates or returns a repository entity based upon its
     * opencode repository id.
     */
    private fun getOrCreate(gitRepository: RepositoryCreateDto): RepositoryEntity {
        logger.info("Checking if repository with ${gitRepository.id} exists")
        var repo = repositoryRepository.findByProjectId(gitRepository.id)
        if (repo == null) {
            logger.info("Creating repository with id ${gitRepository.id}")
            repo = repositoryRepository.save(gitRepository.toDbObject())
        }
        return repo
    }


    /**
     * Queries the gitlab api to get the repo name, url, and id.
     */
    fun getRepositoryInfoFromGitlab(projectId: Long): RepositoryCreateDto {
        return openCodeGitlabApi.getRepositoryInfo(projectId)
    }

    @Transactional(readOnly = true, propagation = Propagation.REQUIRES_NEW)
    fun findByProjectIdAndVisualizationConsentTrue(projectId: Long): RepositoryEntity? {
        return repositoryRepository.findByProjectIdAndVisualizationConsentTrue(projectId)
    }

    @Transactional(readOnly = true, propagation = Propagation.REQUIRES_NEW)
    fun getToolRunByProjectId(projectId: Long): ToolRunDto {
        val toolEntity = toolRunRepository.findFirstByRepository_ProjectIdOrderByCreatedAtDesc(projectId)
        return ToolRunDto.getDtoFromEntity(
            toolEntity
        )
    }

    @Transactional(readOnly = false, propagation = Propagation.REQUIRES_NEW)
    fun updateVisualizationConsent(projectId: Long, repositoryConsentDto: RepositoryConsentDto) {
        val repo = repositoryRepository.findByProjectId(projectId)
        if (repo != null) {
            repo.visualizationConsent = repositoryConsentDto.consent
            logger.info("Updated visualization consent $repositoryConsentDto")
            repositoryRepository.save(repo)
        } else {
            logger.error(
                "Could not update visualization consent because repo" +
                    "was not found $repositoryConsentDto"
            )
        }
    }

    @Transactional(readOnly = true, propagation = Propagation.REQUIRES_NEW)
    fun getAllRepositoriesWithConsent(): List<RepositoryEntity> {
        return repositoryRepository.findByVisualizationConsentTrue()
    }

}
